// Copyright 2008, 2010, 2012-2013 Omni Development, Inc. All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// <http://www.omnigroup.com/developer/sourcecode/sourcelicense/>.

#import "OBTestCase.h"

#import <OmniBase/OmniBase.h>

RCS_ID("$Id$")

#ifdef COVERAGE
#import <crt_externs.h>
#import <mach-o/arch.h>
// When run from Xcode, unit tests get run once for each architecture via various scripts in $SYSTEM_DEVELOPER_DIR/Tools.  This is the easiest way to hook into the per-arch setup.
static void OBTestCaseReportCoverage(void)
{
    const char *coverageTool = getenv("OBCOVERAGE_TOOL");
    const char *archName = getenv("CURRENT_ARCH"); // NXGetLocalArchInfo()->name; --> i486, but we want i386.
    const char *targetName = getenv("OBCOVERAGE_TARGET");
    if (coverageTool) {
        fprintf(stderr, "##\n## Running coverage tool '%s' from '%s' target:%s arch:'%s'...\n##\n", coverageTool, getcwd(NULL, 0), targetName, archName);
        char ***envp = _NSGetEnviron();
        if (envp) {
            char **env = *envp;
            while (*env) {
                fprintf(stderr, "%s\n", *env);
                env++;
            }
        }
        
        int child = fork();
        if (child) {
            // parent -- wait for the child
            int status = 0;
            waitpid(child, &status, WNOHANG);
        } else {
            execle(coverageTool, coverageTool, targetName, archName, NULL, *envp);
            perror("execle");
            exit(1);
        }
    }
}
#endif

// Split out to be a nice breakpoint target.
void _OBReportUnexpectedError(NSError *error)
{
    NSLog(@"Error: %@", [error toPropertyList]);
}

#if defined(OMNI_ASSERTIONS_ON)
static NSString * const OBShouldWaitOnAssertFailureKey = @"OBShouldWaitOnAssertFailure";
#endif

@implementation OBTestCase

#if defined(OMNI_ASSERTIONS_ON)
static void OBTestCaseAssertionHandler(const char *type, const char *expression, const char *file, unsigned int lineNumber, const char *reason)
{
    OBLogAssertionFailure(type, expression, file, lineNumber, reason);
    if ([[NSUserDefaults standardUserDefaults] boolForKey:OBShouldWaitOnAssertFailureKey]) {
        // If we are running unit tests, abort on assertion failure.  We could make assertions throw exceptions, but note that this wouldn't catch cases where you are using 'shouldRaise' and hit an assertion.
#ifdef DEBUG
        // If we're failing in a debug build, give the developer a little time to connect in gdb before crashing
        NSTimeInterval timeToWait = 15.0;
        const char *env = getenv("OBASSERT_TIME_TO_WAIT");
        if (env)
            timeToWait = strtod(env, NULL);
        
        if (timeToWait > 0) {
            fprintf(stderr, "You have %g seconds to attach to pid %u in gdb...\n", timeToWait, getpid());
            [NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:timeToWait]];
        }
#endif
    }
    
    abort();
}
#endif

+ (void)initialize;
{
    OBINITIALIZE;
#ifdef COVERAGE
    atexit(OBTestCaseReportCoverage);
#endif
    
    // Let OBObject get set up (which might run OBPostLoader if we are on the right platform).
    [OBObject class];
    
#if defined(OMNI_ASSERTIONS_ON)
    OBSetAssertionFailureHandler(OBTestCaseAssertionHandler);
    
    @autoreleasepool {
        NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
        NSDictionary *assertionDefaults = [NSDictionary dictionaryWithObjectsAndKeys:
                                           @NO, OBShouldWaitOnAssertFailureKey,
                                           nil];
        [defaults registerDefaults:assertionDefaults];
    }

#endif
}

+ (BOOL)shouldRunSlowUnitTests;
{
    return getenv("RunSlowUnitTests") != NULL;
}

@end
